import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvAdaptive(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvAdaptive'
    bl_label = 'LB Adaptive Comfort'
    sv__out_temp: StringProperty(name='_out_temp', update=updateNode)
    sv__air_temp: StringProperty(name='_air_temp', update=updateNode)
    sv__mrt_: StringProperty(name='_mrt_', update=updateNode)
    sv__air_speed_: StringProperty(name='_air_speed_', update=updateNode)
    sv_adapt_par_: StringProperty(name='adapt_par_', update=updateNode)
    sv__run: StringProperty(name='_run', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_out_temp').prop_name = 'sv__out_temp'
        self.inputs.new('SvStringsSocket', '_air_temp').prop_name = 'sv__air_temp'
        self.inputs.new('SvStringsSocket', '_mrt_').prop_name = 'sv__mrt_'
        self.inputs.new('SvStringsSocket', '_air_speed_').prop_name = 'sv__air_speed_'
        self.inputs.new('SvStringsSocket', 'adapt_par_').prop_name = 'sv_adapt_par_'
        self.inputs.new('SvStringsSocket', '_run').prop_name = 'sv__run'
        self.sv_output_names = ['prevail_temp', 'neutral_temp', 'deg_neutral', 'comfort', 'condition', 'comf_obj']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['prevail_temp', 'neutral_temp', 'deg_neutral', 'comfort', 'condition', 'comf_obj']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_out_temp', '_air_temp', '_mrt_', '_air_speed_', 'adapt_par_', '_run']
        self.sv_input_types = ['System.Object', 'System.Object', 'System.Object', 'System.Object', 'System.Object', 'bool']
        self.sv_input_defaults = [None, None, None, None, None, None]
        self.sv_input_access = ['item', 'item', 'item', 'item', 'item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _out_temp, _air_temp, _mrt_, _air_speed_, adapt_par_, _run):

        try:
            from ladybug.datatype.temperature import Temperature
            from ladybug.datacollection import BaseCollection
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_comfort.collection.adaptive import Adaptive
            from ladybug_comfort.parameter.adaptive import AdaptiveParameter
            from ladybug_comfort.adaptive import t_operative, \
                adaptive_comfort_ashrae55, adaptive_comfort_en15251, \
                cooling_effect_ashrae55, cooling_effect_en15251, \
                adaptive_comfort_conditioned
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_comfort:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        def extract_collections(input_list):
            """Process inputs into collections and floats."""
            defaults = [None, None, _air_temp, 0.1]
            data_colls = []
            for i, input in enumerate(input_list):
                if input is None:
                    input_list[i] = defaults[i]
                elif isinstance(input, BaseCollection):
                    data_colls.append(input)
                else:
                    try:
                        input_list[i] = float(input)
                    except ValueError as e:
                        raise TypeError('input {} is not valid. Expected float or '
                                        'DataCollection. Got {}'.format(input, type(input)))
            return input_list, data_colls
        
        if all_required_inputs(ghenv.Component) and _run is True:
            # Process inputs and assign defaults.
            input_list = [_out_temp, _air_temp, _mrt_, _air_speed_]
            input, data_colls = extract_collections(input_list)
            adapt_par = adapt_par_ or AdaptiveParameter()
        
            if data_colls == []:
                # The inputs are all individual values.
                prevail_temp = input[0]
                to = t_operative(input[1], float(input[2]))
                
                # Determine the ralationship to the neutral temperature
                if adapt_par.conditioning != 0:
                    comf_result = adaptive_comfort_conditioned(prevail_temp, to,
                        adapt_par.conditioning, adapt_par.standard)
                elif adapt_par.ashrae55_or_en15251 is True:
                    comf_result = adaptive_comfort_ashrae55(prevail_temp, to)
                else:
                    comf_result = adaptive_comfort_en15251(prevail_temp, to)
                
                # Determine the cooling effect
                if adapt_par.discrete_or_continuous_air_speed is True:
                    ce = cooling_effect_ashrae55(input[3], to)
                else:
                    ce = cooling_effect_en15251(input[3], to)
                
                # Output results
                neutral_temp = comf_result['t_comf']
                deg_neutral = comf_result['deg_comf']
                comfort = adapt_par.is_comfortable(comf_result, ce)
                condition = adapt_par.thermal_condition(comf_result, ce)
            else:
                # The inputs include Data Collections.
                if not isinstance(_air_temp, BaseCollection):
                    _air_temp = data_colls[0].get_aligned_collection(
                        float(_air_temp), Temperature(), 'C')
                
                comf_obj = Adaptive.from_air_and_rad_temp(_out_temp, _air_temp, _mrt_,
                                                         _air_speed_, adapt_par)
                prevail_temp = comf_obj.prevailing_outdoor_temperature
                neutral_temp = comf_obj.neutral_temperature
                deg_neutral = comf_obj.degrees_from_neutral
                comfort = comf_obj.is_comfortable
                condition = comf_obj.thermal_condition

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvAdaptive)

def unregister():
    bpy.utils.unregister_class(SvAdaptive)
