import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvConstrHeader(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvConstrHeader'
    bl_label = 'LB Construct Header'
    sv__data_type: StringProperty(name='_data_type', update=updateNode)
    sv__unit_: StringProperty(name='_unit_', update=updateNode)
    sv__a_period_: StringProperty(name='_a_period_', update=updateNode)
    sv_metadata_: StringProperty(name='metadata_', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_data_type').prop_name = 'sv__data_type'
        self.inputs.new('SvStringsSocket', '_unit_').prop_name = 'sv__unit_'
        self.inputs.new('SvStringsSocket', '_a_period_').prop_name = 'sv__a_period_'
        self.inputs.new('SvStringsSocket', 'metadata_').prop_name = 'sv_metadata_'
        self.sv_output_names = ['header']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['header']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_data_type', '_unit_', '_a_period_', 'metadata_']
        self.sv_input_types = ['System.Object', 'string', 'System.Object', 'string']
        self.sv_input_defaults = [None, None, None, None]
        self.sv_input_access = ['item', 'item', 'item', 'list']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _data_type, _unit_, _a_period_, metadata_):

        try:
            import ladybug.datatype
            from ladybug.datatype.base import DataTypeBase
            from ladybug.header import Header
            from ladybug.analysisperiod import AnalysisPeriod
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        # error message if the data type is not recognized
        msg = 'The connected _data_type is not recognized.\nMake your own with ' \
            'the "LB Construct Data Type" component or choose from the following:' \
            '\n{}'.format('\n'.join(ladybug.datatype.BASETYPES))
        
        
        if all_required_inputs(ghenv.Component):
            if isinstance(_data_type, DataTypeBase):
                pass
            elif isinstance(_data_type, str):
                _data_type = _data_type.title().replace(' ', '')
                if _data_type not in ladybug.datatype.TYPES:
                    raise TypeError(msg)
                _data_type = ladybug.datatype.TYPESDICT[_data_type]()
            else:
                raise TypeError(msg)
        
            if _unit_ is None:
                _unit_ = _data_type.units[0]
        
            if _a_period_ is None:
                _a_period_ = AnalysisPeriod()
        
            metadata_dict = {}
            if metadata_ != []:
                for prop in metadata_:
                    key, value = prop.split(':')
                    metadata_dict[key] = value.strip()
        
            header = Header(_data_type, _unit_, _a_period_, metadata_dict)

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvConstrHeader)

def unregister():
    bpy.utils.unregister_class(SvConstrHeader)
