import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvToStep(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvToStep'
    bl_label = 'LB Convert to Timestep'
    sv__data: StringProperty(name='_data', update=updateNode)
    sv__timestep_: StringProperty(name='_timestep_', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_data').prop_name = 'sv__data'
        self.inputs.new('SvStringsSocket', '_timestep_').prop_name = 'sv__timestep_'
        self.sv_output_names = ['data']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['data']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_data', '_timestep_']
        self.sv_input_types = ['System.Object', 'int']
        self.sv_input_defaults = [None, None]
        self.sv_input_access = ['item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _data, _timestep_):

        try:
            from ladybug.datacollection import HourlyDiscontinuousCollection, \
                HourlyContinuousCollection
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            # check the inputs
            assert isinstance(_data, HourlyDiscontinuousCollection), \
                ' Connected _data is not a Hourly Data Collection. Got{}'.format(type(_data))
            if _timestep_ is not None:
                valid_timesteps = _data.header.analysis_period.VALIDTIMESTEPS
                assert _timestep_ in valid_timesteps, ' Connected _timestep_ is not a'\
                    ' hourly timestep.\n Got {}. Choose from: {}'.format(
                    _timestep_, sorted(valid_timesteps.keys()))
        
            # if the data is not continuous, interpolate over holes.
            if not isinstance(_data, HourlyContinuousCollection):
                if _data.validated_a_period is False:
                    _data = data.validate_analysis_period
                _data = _data.interpolate_holes()
        
            # convert the data to the timestep
            if _timestep_ and _timestep_ != _data.header.analysis_period.timestep:
                if _timestep_ < _data.header.analysis_period.timestep:
                    data = _data.cull_to_timestep(_timestep_)
                else:
                    data = _data.interpolate_to_timestep(_timestep_)
            else:
                data = _data

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvToStep)

def unregister():
    bpy.utils.unregister_class(SvToStep)
