import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvCreateLegend(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvCreateLegend'
    bl_label = 'LB Create Legend'
    sv__values: StringProperty(name='_values', update=updateNode)
    sv__base_plane_: StringProperty(name='_base_plane_', update=updateNode)
    sv_title_: StringProperty(name='title_', update=updateNode)
    sv_legend_par_: StringProperty(name='legend_par_', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_values').prop_name = 'sv__values'
        self.inputs.new('SvStringsSocket', '_base_plane_').prop_name = 'sv__base_plane_'
        self.inputs.new('SvStringsSocket', 'title_').prop_name = 'sv_title_'
        self.inputs.new('SvStringsSocket', 'legend_par_').prop_name = 'sv_legend_par_'
        self.sv_output_names = ['mesh', 'title_obj', 'label_objs', 'label_text', 'colors']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['mesh', 'title_obj', 'label_objs', 'label_text', 'colors']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_values', '_base_plane_', 'title_', 'legend_par_']
        self.sv_input_types = ['double', 'Plane', 'string', 'System.Object']
        self.sv_input_defaults = [None, None, None, None]
        self.sv_input_access = ['list', 'item', 'item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _values, _base_plane_, title_, legend_par_):

        
        try:
            from ladybug.legend import Legend, LegendParameters
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.togeometry import to_plane
            from ladybug_tools.fromobjects import legend_objects
            from ladybug_tools.color import color_to_color
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            # set default values
            legend_par_ = legend_par_.duplicate() if legend_par_ is not None else \
                LegendParameters()
            if _base_plane_:
                legend_par_.base_plane = to_plane(_base_plane_)
            legend_par_.title = title_
        
            # create the legend
            legend = Legend(_values, legend_par_)
        
            # separate all of the outputs from this component
            tools_objs = legend_objects(legend)
            mesh = tools_objs[0]
            title_obj = tools_objs[1]
            label_objs = tools_objs[2:]
            label_text = legend.segment_text
            colors = [color_to_color(col) for col in legend.value_colors]
        

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvCreateLegend)

def unregister():
    bpy.utils.unregister_class(SvCreateLegend)
