import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvSkyMatrix(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvSkyMatrix'
    bl_label = 'LB Cumulative Sky Matrix'
    sv_north_: StringProperty(name='north_', update=updateNode)
    sv__location: StringProperty(name='_location', update=updateNode)
    sv__direct_rad: StringProperty(name='_direct_rad', update=updateNode)
    sv__diffuse_rad: StringProperty(name='_diffuse_rad', update=updateNode)
    sv__hoys_: StringProperty(name='_hoys_', update=updateNode)
    sv_high_density_: StringProperty(name='high_density_', update=updateNode)
    sv__folder_: StringProperty(name='_folder_', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', 'north_').prop_name = 'sv_north_'
        self.inputs.new('SvStringsSocket', '_location').prop_name = 'sv__location'
        self.inputs.new('SvStringsSocket', '_direct_rad').prop_name = 'sv__direct_rad'
        self.inputs.new('SvStringsSocket', '_diffuse_rad').prop_name = 'sv__diffuse_rad'
        self.inputs.new('SvStringsSocket', '_hoys_').prop_name = 'sv__hoys_'
        self.inputs.new('SvStringsSocket', 'high_density_').prop_name = 'sv_high_density_'
        self.inputs.new('SvStringsSocket', '_folder_').prop_name = 'sv__folder_'
        self.sv_output_names = ['sky_mtx']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['sky_mtx']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['north_', '_location', '_direct_rad', '_diffuse_rad', '_hoys_', 'high_density_', '_folder_']
        self.sv_input_types = ['System.Object', 'System.Object', 'System.Object', 'System.Object', 'int', 'bool', 'string']
        self.sv_input_defaults = [None, None, None, None, None, None, None]
        self.sv_input_access = ['item', 'item', 'item', 'item', 'list', 'item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, north_, _location, _direct_rad, _diffuse_rad, _hoys_, high_density_, _folder_):

        import os
        import subprocess
        import math
        
        try:
            from ladybug_geometry.geometry2d.pointvector import Vector2D
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))
        
        try:
            from ladybug.wea import Wea
            from ladybug.viewsphere import view_sphere
            from ladybug.dt import DateTime
            from ladybug.config import folders as lb_folders
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.togeometry import to_vector2d
            from ladybug_tools.sverchok import all_required_inputs, objectify_output
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        # TODO: Remove dependency on honeybee + Radiance after genskymtx is in its own LB extension
        try:
            from honeybee_radiance.config import folders as hb_folders
            compatible_rad_date = (2020, 9, 3)
            hb_url = 'https://github.com/ladybug-tools/lbt-sverchok/wiki/1.4-Compatibility-Matrix'
            rad_msg = 'Download and install the version of Radiance listed in the Ladybug ' \
                'Tools compatibility matrix\n{}'.format(hb_url)
            assert hb_folders.radiance_path is not None, \
                'No Radiance installation was found on this machine.\n{}'.format(rad_msg)
            assert hb_folders.radiance_version_date >= compatible_rad_date, \
                'The installed Radiance is not from {} or later.' \
                '\n{}'.format('/'.join(str(v) for v in compatible_rad_date), rad_msg)
            # get the path to the gemdaymtx executable
            gendaymtx_exe = os.path.join(hb_folders.radbin_path, 'gendaymtx.exe') if \
                os.name == 'nt' else os.path.join(hb_folders.radbin_path, 'gendaymtx')
        except ImportError as e:
            raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))
        
        
        # constants for converting RGB values output by gendaymtx to broadband radiation
        PATCHES_PER_ROW = {
            1: view_sphere.TREGENZA_PATCHES_PER_ROW + (1,),
            2: view_sphere.REINHART_PATCHES_PER_ROW + (1,)
        }
        PATCH_ROW_COEFF = {
            1 : view_sphere.TREGENZA_COEFFICIENTS,
            2: view_sphere.REINHART_COEFFICIENTS
        }
        
        
        def broadband_radiation(patch_row_str, row_number, wea_duration, sky_density=1):
            """Parse a row of gendaymtx RGB patch data in W/sr/m2 to radiation in kWh/m2.
        
            This includes aplying broadband weighting to the RGB bands, multiplication
            by the steradians of each patch, and multiplying by the duration of time that
            they sky matrix represents in hours.
        
            Args:
                patch_row_str: Text string for a single row of RGB patch data.
                row_number: Interger for the row number that the patch corresponds to.
                sky_density: Integer (either 1 or 2) for the density.
                wea_duration: Number for the duration of the Wea in hours. This is used
                    to convert between the average value output by the command and the
                    cumulative value that is needed for all ladybug analyses.
            """
            R, G, B = patch_row_str.split(' ')
            weight_val = 0.265074126 * float(R) + 0.670114631 * float(G) + 0.064811243 * float(B)
            return weight_val * PATCH_ROW_COEFF[sky_density][row_number] * wea_duration / 1000
        
        
        def parse_mtx_data(data_str, wea_duration, sky_density=1):
            """Parse a string of Radiance gendaymtx data to a list of radiation-per-patch.
        
            This function handles the removing of the header and the conversion of the
            RGB irradianc-=per-steraidian values to broadband radiation. It also removes
            the first patch, which is the ground and is not used by Ladybug.
        
            Args:
                data_str: The string that has been output by gendaymtx to stdout.
                wea_duration: Number for the duration of the Wea in hours. This is used
                    to convert between the average value output by the command and the
                    cumulative value that is needed for all ladybug analyses.
                sky_density: Integer (either 1 or 2) for the density.
            """
            # split lines and remove the header, ground patch and last line break
            data_lines = data_str.split('\n')
            patch_lines = data_lines[9:-1]
        
            # loop through the rows and convert the radiation RGB values
            broadband_irr = []
            patch_counter = 0
            for i, row_patch_count in enumerate(PATCHES_PER_ROW[sky_density]):
                row_slice = patch_lines[patch_counter:patch_counter + row_patch_count]
                irr_vals = (broadband_radiation(row, i, wea_duration, sky_density)
                            for row in row_slice)
                broadband_irr.extend(irr_vals)
                patch_counter += row_patch_count
            return broadband_irr
        
        
        if all_required_inputs(ghenv.Component):
            # process and set defaults for all of the global inputs
            if north_ is not None:  # process the north_
                try:
                    north_ = math.degrees(
                        to_vector2d(north_).angle_clockwise(Vector2D(0, 1)))
                except AttributeError:  # north angle instead of vector
                    north_ = float(north_)
            else:
                north_ = 0
            density = 2 if high_density_ else 1
        
            # filter the radiation by _hoys if they are input
            if len(_hoys_) != 0:
                _direct_rad = _direct_rad.filter_by_hoys(_hoys_)
                _diffuse_rad = _diffuse_rad.filter_by_hoys(_hoys_)
        
            # create the wea and write it to the default_epw_folder
            wea = Wea(_location, _direct_rad, _diffuse_rad)
            wea_duration = len(wea) / wea.timestep
            wea_folder = _folder_ if _folder_ is not None else \
                os.path.join(lb_folders.default_epw_folder, 'sky_matrices')
            metd = _direct_rad.header.metadata
            wea_basename = os.path.join(wea_folder, metd['city'].replace(' ', '_')) \
                if 'city' in metd else 'unnamed'
            wea_file =wea.write(wea_basename)
        
            # execute the Radiance gendaymtx command
            use_shell = True if os.name == 'nt' else False
            # command for direct patches
            cmds = [gendaymtx_exe, '-m', str(density), '-d', '-O1', '-A', wea_file]
            process = subprocess.Popen(cmds, stdout=subprocess.PIPE, shell=use_shell)
            stdout = process.communicate()
            dir_data_str = stdout[0]
            # command for diffuse patches
            cmds = [gendaymtx_exe, '-m', str(density), '-s', '-O1', '-A', wea_file]
            process = subprocess.Popen(cmds, stdout=subprocess.PIPE, shell=use_shell)
            stdout = process.communicate()
            diff_data_str = stdout[0]
        
            # parse the data into a single matrix
            dir_vals = parse_mtx_data(dir_data_str, wea_duration, density)
            diff_vals = parse_mtx_data(diff_data_str, wea_duration, density)
        
            # collect sky metadata like the north, which will be used by other components
            metadata = [north_]
            if _hoys_:
                metadata.extend([DateTime.from_hoy(h) for h in (_hoys_[0], _hoys_[-1])])
            else:
                metadata.extend([wea.analysis_period.st_time, wea.analysis_period.end_time])
            for key, val in list(_direct_rad.header.metadata.items()):
                metadata.append('{} : {}'.format(key, val))
        
            # wrap everything together into an object to output from the component
            mtx_data = (metadata, dir_vals, diff_vals)
            sky_mtx = objectify_output('Cumulative Sky Matrix', mtx_data)
        

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvSkyMatrix)

def unregister():
    bpy.utils.unregister_class(SvSkyMatrix)
