import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvDownloadEPW(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvDownloadEPW'
    bl_label = 'LB Download Weather'
    sv__weather_URL: StringProperty(name='_weather_URL', update=updateNode)
    sv__folder_: StringProperty(name='_folder_', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_weather_URL').prop_name = 'sv__weather_URL'
        self.inputs.new('SvStringsSocket', '_folder_').prop_name = 'sv__folder_'
        self.sv_output_names = ['epw_file', 'stat_file', 'ddy_file']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['epw_file', 'stat_file', 'ddy_file']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_weather_URL', '_folder_']
        self.sv_input_types = ['string', 'string']
        self.sv_input_defaults = [None, None]
        self.sv_input_access = ['item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _weather_URL, _folder_):

        try:
            from ladybug.futil import unzip_file
            from ladybug.config import folders
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.download import download_file
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        import os
        
        
        if all_required_inputs(ghenv.Component):
            # name for the weather files
            if _weather_URL.lower().endswith('.zip'):
                # onebuilding URL type
                _folder_name = _weather_URL.split('/')[-1][:-4]
            else:
                # dept of energy URL type
                _folder_name = _weather_URL.split('/')[-2]
            
            # create default working_dir
            if _folder_ is None:
                _folder_ = folders.default_epw_folder
            try:
                _folder_.decode('ascii')
            except UnicodeDecodeError:
                raise UnicodeDecodeError(
                    '\nYour download folder "{}" contains non-ASCII characters\n'
                    '\nUse the _folder_ input to this component to download EPW files'
                    ' to a valid location.'.format(_folder_))
            else:
                print('Files will be downloaded to: {}'.format(_folder_))
            
            
            # default file names
            epw = os.path.join(_folder_, _folder_name, _folder_name + '.epw')
            stat = os.path.join(_folder_, _folder_name, _folder_name + '.stat')
            ddy = os.path.join(_folder_, _folder_name, _folder_name + '.ddy')
            
            # download and unzip the files if they do not exist
            if not os.path.isfile(epw) or not os.path.isfile(stat) or not os.path.isfile(ddy):
                zip_file_path = os.path.join(_folder_, _folder_name, _folder_name + '.zip')
                download_file(_weather_URL, zip_file_path, True)
                unzip_file(zip_file_path)
            
            # set output
            epw_file, stat_file, ddy_file = epw, stat, ddy

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvDownloadEPW)

def unregister():
    bpy.utils.unregister_class(SvDownloadEPW)
