import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvEPWtoDDY(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvEPWtoDDY'
    bl_label = 'LB EPW to DDY'
    sv__epw_file: StringProperty(name='_epw_file', update=updateNode)
    sv__percentile_: StringProperty(name='_percentile_', update=updateNode)
    sv__folder_: StringProperty(name='_folder_', update=updateNode)
    sv__write: StringProperty(name='_write', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_epw_file').prop_name = 'sv__epw_file'
        self.inputs.new('SvStringsSocket', '_percentile_').prop_name = 'sv__percentile_'
        self.inputs.new('SvStringsSocket', '_folder_').prop_name = 'sv__folder_'
        self.inputs.new('SvStringsSocket', '_write').prop_name = 'sv__write'
        self.sv_output_names = ['ddy_file']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['ddy_file']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_epw_file', '_percentile_', '_folder_', '_write']
        self.sv_input_types = ['string', 'double', 'string', 'bool']
        self.sv_input_defaults = [None, None, None, None]
        self.sv_input_access = ['item', 'item', 'item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _epw_file, _percentile_, _folder_, _write):

        import os
        
        try:
            from ladybug.epw import EPW
            from ladybug.config import folders
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component) and _write:
            # set default values
            _percentile_ = 0.4 if _percentile_ is None else _percentile_
            _folder_ = os.path.join(folders.default_epw_folder, 'ddy') if _folder_ \
                is None else _folder_
            f_name = os.path.basename(_epw_file).replace('.epw', '.ddy')
            f_path = os.path.join(_folder_, f_name)
        
            # create the DDY file
            epw = EPW(_epw_file)
            ddy_file = epw.to_ddy(f_path, _percentile_)
        

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvEPWtoDDY)

def unregister():
    bpy.utils.unregister_class(SvEPWtoDDY)
