import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvGenPts(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvGenPts'
    bl_label = 'LB Generate Point Grid'
    sv__geometry: StringProperty(name='_geometry', update=updateNode)
    sv__grid_size: StringProperty(name='_grid_size', update=updateNode)
    sv__offset_dist_: StringProperty(name='_offset_dist_', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_geometry').prop_name = 'sv__geometry'
        self.inputs.new('SvStringsSocket', '_grid_size').prop_name = 'sv__grid_size'
        self.inputs.new('SvStringsSocket', '_offset_dist_').prop_name = 'sv__offset_dist_'
        self.sv_output_names = ['points', 'vectors', 'face_areas', 'mesh']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['points', 'vectors', 'face_areas', 'mesh']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_geometry', '_grid_size', '_offset_dist_']
        self.sv_input_types = ['GeometryBase', 'double', 'double']
        self.sv_input_defaults = [None, None, None]
        self.sv_input_access = ['item', 'item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _geometry, _grid_size, _offset_dist_):

        try:
            from ladybug_tools.togeometry import to_gridded_mesh3d, to_mesh3d
            from ladybug_tools.fromgeometry import from_mesh3d, from_point3d, from_vector3d
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            # check the input and generate the mesh.
            _offset_dist_ = _offset_dist_ or 0
            try:  # assume it's a Blender Ladybug Brep
                lb_mesh = to_gridded_mesh3d(_geometry, _grid_size, _offset_dist_)
            except TypeError:  # assume it's a Blender Ladybug Mesh
                try:
                    lb_mesh = to_mesh3d(_geometry)
                except TypeError:  # unidientified geometry type
                    raise TypeError(
                        '_geometry must be a Brep or a Mesh. Got {}.'.format(type(_geometry)))
            
            # generate the test points, vectors, and areas.
            points = [from_point3d(pt) for pt in lb_mesh.face_centroids]
            vectors = [from_vector3d(vec) for vec in lb_mesh.face_normals]
            face_areas = lb_mesh.face_areas
            mesh = from_mesh3d(lb_mesh)

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvGenPts)

def unregister():
    bpy.utils.unregister_class(SvGenPts)
