import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvHourlyPlot(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvHourlyPlot'
    bl_label = 'LB Hourly Plot'
    sv__data: StringProperty(name='_data', update=updateNode)
    sv__base_pt_: StringProperty(name='_base_pt_', update=updateNode)
    sv__x_dim_: StringProperty(name='_x_dim_', update=updateNode)
    sv__y_dim_: StringProperty(name='_y_dim_', update=updateNode)
    sv__z_dim_: StringProperty(name='_z_dim_', update=updateNode)
    sv_reverse_y_: StringProperty(name='reverse_y_', update=updateNode)
    sv_legend_par_: StringProperty(name='legend_par_', update=updateNode)
    sv_statement_: StringProperty(name='statement_', update=updateNode)
    sv_period_: StringProperty(name='period_', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_data').prop_name = 'sv__data'
        self.inputs.new('SvStringsSocket', '_base_pt_').prop_name = 'sv__base_pt_'
        self.inputs.new('SvStringsSocket', '_x_dim_').prop_name = 'sv__x_dim_'
        self.inputs.new('SvStringsSocket', '_y_dim_').prop_name = 'sv__y_dim_'
        self.inputs.new('SvStringsSocket', '_z_dim_').prop_name = 'sv__z_dim_'
        self.inputs.new('SvStringsSocket', 'reverse_y_').prop_name = 'sv_reverse_y_'
        self.inputs.new('SvStringsSocket', 'legend_par_').prop_name = 'sv_legend_par_'
        self.inputs.new('SvStringsSocket', 'statement_').prop_name = 'sv_statement_'
        self.inputs.new('SvStringsSocket', 'period_').prop_name = 'sv_period_'
        self.sv_output_names = ['mesh', 'legend', 'borders', 'labels', 'title']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['mesh', 'legend', 'borders', 'labels', 'title']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_data', '_base_pt_', '_x_dim_', '_y_dim_', '_z_dim_', 'reverse_y_', 'legend_par_', 'statement_', 'period_']
        self.sv_input_types = ['System.Object', 'Point3d', 'double', 'double', 'double', 'bool', 'System.Object', 'string', 'System.Object']
        self.sv_input_defaults = [None, None, None, None, None, None, None, None, None]
        self.sv_input_access = ['list', 'item', 'item', 'item', 'item', 'item', 'list', 'item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _data, _base_pt_, _x_dim_, _y_dim_, _z_dim_, reverse_y_, legend_par_, statement_, period_):

        try:
            from ladybug.datacollection import HourlyContinuousCollection
            from ladybug.hourlyplot import HourlyPlot
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_geometry.geometry3d.pointvector import Point3D
            from ladybug_geometry.geometry3d.plane import Plane
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.config import conversion_to_meters
            from ladybug_tools.togeometry import to_point3d
            from ladybug_tools.fromgeometry import from_mesh3d, from_mesh2d, \
                from_polyline2d, from_linesegment2d
            from ladybug_tools.text import text_objects
            from ladybug_tools.fromobjects import legend_objects
            from ladybug_tools.sverchok import all_required_inputs, list_to_data_tree
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            # apply any analysis periods and conditional statements to the input collections
            if period_ is not None:
                _data = [coll.filter_by_analysis_period(period_) for coll in _data]
            if statement_ is not None:
                _data = HourlyContinuousCollection.filter_collections_by_statement(
                    _data, statement_)
        
            # set default values for the chart dimensions
            _base_pt_ = to_point3d(_base_pt_) if _base_pt_ is not None else Point3D()
            _x_dim_ = _x_dim_ if _x_dim_ is not None else 1.0 / conversion_to_meters()
            _y_dim_ = _y_dim_ if _y_dim_ is not None else 4.0 / conversion_to_meters()
            _z_dim_ = _z_dim_ if _z_dim_ is not None else 0
            reverse_y_ = reverse_y_ if reverse_y_ is not None else False
        
            # empty lists of objects to be filled with visuals
            mesh, title, all_legends, all_borders, all_labels = [], [], [], [], []
        
            for i, data_coll in enumerate(_data):
                try:  # sense when several legend parameters are connected
                    lpar = legend_par_[i]
                except IndexError:
                    lpar = None if len(legend_par_) == 0 else legend_par_[-1]
        
                # create the hourly plot object and get the main pieces of geometry
                hour_plot = HourlyPlot(data_coll, lpar, _base_pt_,
                                       _x_dim_, _y_dim_, _z_dim_, reverse_y_)
                msh = from_mesh2d(hour_plot.colored_mesh2d, _base_pt_.z) if _z_dim_ == 0 else \
                    from_mesh3d(hour_plot.colored_mesh3d)
                mesh.append(msh)
                border = [from_polyline2d(hour_plot.chart_border2d, _base_pt_.z)] + \
                    [from_linesegment2d(line, _base_pt_.z) for line in hour_plot.hour_lines2d] + \
                    [from_linesegment2d(line, _base_pt_.z) for line in hour_plot.month_lines2d]
                all_borders.append(border)
                legnd = legend_objects(hour_plot.legend)
                all_legends.append(legnd)
                tit_txt = text_objects(hour_plot.title_text, hour_plot.lower_title_location,
                                       hour_plot.legend_parameters.text_height,
                                       hour_plot.legend_parameters.font)
                title.append(tit_txt)
        
                # create the text label objects
                label1 = [text_objects(txt, Plane(o=Point3D(pt.x, pt.y, _base_pt_.z)),
                                       hour_plot.legend_parameters.text_height,
                                       hour_plot.legend_parameters.font, 2, 3)
                          for txt, pt in zip(hour_plot.hour_labels, hour_plot.hour_label_points2d)]
                label2 = [text_objects(txt, Plane(o=Point3D(pt.x, pt.y, _base_pt_.z)),
                                       hour_plot.legend_parameters.text_height,
                                       hour_plot.legend_parameters.font, 1, 0)
                          for txt, pt in zip(hour_plot.month_labels, hour_plot.month_label_points2d)]
                all_labels.append(label1 + label2)
                
                # increment the base point so that the next chart doesn't overlap this one
                try:
                    next_tstep = _data[i + 1].header.analysis_period.timestep
                except IndexError:
                    next_tstep = 1
                increment = 24 * next_tstep * _y_dim_ * 1.5
                _base_pt_ = Point3D(_base_pt_.x, _base_pt_.y - increment, _base_pt_.z)
        
            # convert nexted lists into data trees
            legend = list_to_data_tree(all_legends)
            borders = list_to_data_tree(all_borders)
            labels = list_to_data_tree(all_labels)
        

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvHourlyPlot)

def unregister():
    bpy.utils.unregister_class(SvHourlyPlot)
