import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvHumanToSky(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvHumanToSky'
    bl_label = 'LB Human to Sky Relation'
    sv_north_: StringProperty(name='north_', update=updateNode)
    sv__location: StringProperty(name='_location', update=updateNode)
    sv__position: StringProperty(name='_position', update=updateNode)
    sv__context: StringProperty(name='_context', update=updateNode)
    sv__pt_count_: StringProperty(name='_pt_count_', update=updateNode)
    sv__height_: StringProperty(name='_height_', update=updateNode)
    sv_parallel_: StringProperty(name='parallel_', update=updateNode)
    sv__run: StringProperty(name='_run', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', 'north_').prop_name = 'sv_north_'
        self.inputs.new('SvStringsSocket', '_location').prop_name = 'sv__location'
        self.inputs.new('SvStringsSocket', '_position').prop_name = 'sv__position'
        self.inputs.new('SvStringsSocket', '_context').prop_name = 'sv__context'
        self.inputs.new('SvStringsSocket', '_pt_count_').prop_name = 'sv__pt_count_'
        self.inputs.new('SvStringsSocket', '_height_').prop_name = 'sv__height_'
        self.inputs.new('SvStringsSocket', 'parallel_').prop_name = 'sv_parallel_'
        self.inputs.new('SvStringsSocket', '_run').prop_name = 'sv__run'
        self.sv_output_names = ['human_points', 'human_line', 'fract_body_exp', 'sky_exposure']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['human_points', 'human_line', 'fract_body_exp', 'sky_exposure']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['north_', '_location', '_position', '_context', '_pt_count_', '_height_', 'parallel_', '_run']
        self.sv_input_types = ['System.Object', 'System.Object', 'Point3d', 'GeometryBase', 'int', 'double', 'System.Object', 'bool']
        self.sv_input_defaults = [None, None, None, None, None, None, None, None]
        self.sv_input_access = ['item', 'item', 'list', 'list', 'item', 'item', 'item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, north_, _location, _position, _context, _pt_count_, _height_, parallel_, _run):

        try:
            from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
            from ladybug_geometry.geometry3d.line import LineSegment3D
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))
        
        try:
            from ladybug.sunpath import Sunpath
            from ladybug.viewsphere import view_sphere
            from ladybug.datacollection import HourlyContinuousCollection
            from ladybug.header import Header
            from ladybug.analysisperiod import AnalysisPeriod
            from ladybug.datatype.fraction import Fraction
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.config import conversion_to_meters
            from ladybug_tools.togeometry import to_point3d, to_vector3d
            from ladybug_tools.fromgeometry import from_point3d, from_vector3d, \
                from_linesegment3d
            from ladybug_tools.intersect import join_geometry_to_mesh, intersect_mesh_rays
            from ladybug_tools.sverchok import all_required_inputs, list_to_data_tree
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        def human_height_points(position, height, pt_count):
            """Get a list of points and a line representing the human geometry.
        
            Args:
                position: Blender Ladybug point for the position of the human.
                height: Number for the height of the human.
                pt_count: Integer for the number of points representing the human.
        
            Returns:
                 A tuple with human points as first element and human line as second.
                 Both geomtries are Blender Ladybug geometries.
            """
            lb_feet_pt = to_point3d(position).move(Vector3D(0, 0, height / 100))
            lb_hum_line = LineSegment3D(lb_feet_pt, Vector3D(0, 0, height))
            lb_pts = [lb_hum_line.midpoint] if pt_count == 1 else \
                lb_hum_line.subdivide_evenly(pt_count - 1)
            if len(lb_pts) == pt_count - 1:  # sometimes tolerance kills the last point
                lb_pts.append(lb_feet_pt.move(Vector3D(0, 0, height)))
            h_points = [from_point3d(pt) for pt in lb_pts]
            return h_points, from_linesegment3d(lb_hum_line)
        
        
        def fract_exposed_from_mtx(person_sun_int_matrix, day_pattern):
            """Get a Data Collection of fraction exposed values from an intersection matrix.
        
            Args:
                person_sun_int_matrix: An intersection matrix of 0s and 1s for the points
                    of a single person.
                day_pattern: A list of 8760 booleans indicating whether the sun is
                    up (True) or down (Fasle).
        
            Returns:
                 A data collection for the fraction of body exposed.
             """
            pt_count = len(person_sun_int_matrix)
            fract_per_sun = [sum(pt_int_ar) / pt_count for pt_int_ar in zip(*person_sun_int_matrix)]
            fract_exp_vals = []
            per_sun_i = 0
            for is_sun in day_pattern:
                if is_sun:
                    fract_exp_vals.append(fract_per_sun[per_sun_i])
                    per_sun_i += 1
                else:
                    fract_exp_vals.append(0)
            meta_dat = {'type': 'Fraction of Body Exposed to Direct Sun'}
            fract_exp_head = Header(Fraction(), 'fraction', AnalysisPeriod(), meta_dat)
            return HourlyContinuousCollection(fract_exp_head, fract_exp_vals)
        
        
        def sky_exposure_from_mtx(person_sky_int_matrix, patch_weights):
            """Get a the sky exposure from an intersection matrix.
        
            Args:
                person_sky_int_matrix: An intersection matrix of 0s and 1s for the
                    points of a person intersected with the 145 tregenza patches.
                patch_weights: A list of 145 weights to be applies to the patches.
        
            Returns:
                 A value for the sky exposure of the person.
             """
            pt_count = len(person_sky_int_matrix)
            sky_exp_per_pt = [sum((r * w) / 145 for r, w in zip(int_list, patch_wghts))
                              for int_list in person_sky_int_matrix]
            return sum(sky_exp_per_pt) / pt_count
        
        
        if all_required_inputs(ghenv.Component):
            # process the north input if specified
            if north_ is not None:  # process the north_
                try:
                    north_ = math.degrees(to_vector2d(north_).angle_clockwise(Vector2D(0, 1)))
                except AttributeError:  # north angle instead of vector
                    north_ = float(north_)
            else:
                north_ = 0
        
            # set the default point count and height if unspecified
            _pt_count_ = _pt_count_ if _pt_count_ is not None else 1
            _height_ = _height_ if _height_ is not None else 1.8 / conversion_to_meters()
        
            # create the points representing the human geometry
            human_points = []
            human_line = []
            for pos in _position:
                hpts, hlin = human_height_points(pos, _height_, _pt_count_)
                human_points.extend(hpts)
                human_line.append(hlin)
        
            if _run:
                # mesh the context for the intersection calculation
                shade_mesh = join_geometry_to_mesh(_context)
        
                # generate the sun vectors for each sun-up hour of the year
                sp = Sunpath.from_location(_location, north_)
                sun_vecs = []
                day_pattern = []
                for hoy in range(8760):
                    sun = sp.calculate_sun_from_hoy(hoy)
                    day_pattern.append(sun.is_during_day)
                    if sun.is_during_day:
                        sun_vecs.append(from_vector3d(sun.sun_vector_reversed))
        
                # intersect the sun vectors with the context and compute fraction exposed
                sun_int_matrix, angles = intersect_mesh_rays(
                    shade_mesh, human_points, sun_vecs, parallel=parallel_)
                fract_body_exp = []
                for i in range(0, len(human_points), _pt_count_):
                    fract_body_exp.append(
                        fract_exposed_from_mtx(sun_int_matrix[i:i + _pt_count_], day_pattern))
        
                # generate the vectors and weights for sky exposure
                sky_vecs = [from_vector3d(vec) for vec in view_sphere.tregenza_dome_vectors]
                patch_wghts = view_sphere.dome_patch_weights(1)
        
                # compute the sky exposure
                sky_int_matrix, angles = intersect_mesh_rays(
                    shade_mesh, human_points, sky_vecs, parallel=parallel_)
                sky_exposure = []
                for i in range(0, len(human_points), _pt_count_):
                    sky_exposure.append(
                        sky_exposure_from_mtx(sky_int_matrix[i:i + _pt_count_], patch_wghts))
        

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvHumanToSky)

def unregister():
    bpy.utils.unregister_class(SvHumanToSky)
