import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvHumidityR(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvHumidityR'
    bl_label = 'LB Humidity Metrics'
    sv__dry_bulb: StringProperty(name='_dry_bulb', update=updateNode)
    sv__rel_humid: StringProperty(name='_rel_humid', update=updateNode)
    sv__pressure_: StringProperty(name='_pressure_', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_dry_bulb').prop_name = 'sv__dry_bulb'
        self.inputs.new('SvStringsSocket', '_rel_humid').prop_name = 'sv__rel_humid'
        self.inputs.new('SvStringsSocket', '_pressure_').prop_name = 'sv__pressure_'
        self.sv_output_names = ['humid_ratio', 'enthalpy', 'wet_bulb', 'dew_point']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['humid_ratio', 'enthalpy', 'wet_bulb', 'dew_point']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_dry_bulb', '_rel_humid', '_pressure_']
        self.sv_input_types = ['System.Object', 'System.Object', 'System.Object']
        self.sv_input_defaults = [None, None, None]
        self.sv_input_access = ['item', 'item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _dry_bulb, _rel_humid, _pressure_):

        try:
            from ladybug.psychrometrics import humid_ratio_from_db_rh, enthalpy_from_db_hr, \
                wet_bulb_from_db_rh, dew_point_from_db_rh
            from ladybug.datacollection import HourlyContinuousCollection
            from ladybug.datatype.fraction import HumidityRatio
            from ladybug.datatype.specificenergy import Enthalpy
            from ladybug.datatype.temperature import WetBulbTemperature, DewPointTemperature
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            if _pressure_ is None:
                _pressure_ = 101325
        
            humid_ratio = HourlyContinuousCollection.compute_function_aligned(
                humid_ratio_from_db_rh, [_dry_bulb, _rel_humid, _pressure_],
                HumidityRatio(), 'fraction')
            
            enthalpy = HourlyContinuousCollection.compute_function_aligned(
                enthalpy_from_db_hr, [_dry_bulb, humid_ratio], Enthalpy(), 'kJ/kg')
        
            wet_bulb = HourlyContinuousCollection.compute_function_aligned(
                wet_bulb_from_db_rh, [_dry_bulb, _rel_humid, _pressure_],
                WetBulbTemperature(), 'C')
        
            dew_point = HourlyContinuousCollection.compute_function_aligned(
                dew_point_from_db_rh, [_dry_bulb, _rel_humid],
                DewPointTemperature(), 'C')

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvHumidityR)

def unregister():
    bpy.utils.unregister_class(SvHumidityR)
