import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvImportEPW(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvImportEPW'
    bl_label = 'LB Import EPW'
    sv__epw_file: StringProperty(name='_epw_file', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_epw_file').prop_name = 'sv__epw_file'
        self.sv_output_names = ['location', 'dry_bulb_temperature', 'dew_point_temperature', 'relative_humidity', 'wind_speed', 'wind_direction', 'direct_normal_rad', 'diffuse_horizontal_rad', 'global_horizontal_rad', 'horizontal_infrared_rad', 'direct_normal_ill', 'diffuse_horizontal_ill', 'global_horizontal_ill', 'total_sky_cover', 'barometric_pressure', 'model_year', 'ground_temperature']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['location', 'dry_bulb_temperature', 'dew_point_temperature', 'relative_humidity', 'wind_speed', 'wind_direction', 'direct_normal_rad', 'diffuse_horizontal_rad', 'global_horizontal_rad', 'horizontal_infrared_rad', 'direct_normal_ill', 'diffuse_horizontal_ill', 'global_horizontal_ill', 'total_sky_cover', 'barometric_pressure', 'model_year', 'ground_temperature']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_epw_file']
        self.sv_input_types = ['string']
        self.sv_input_defaults = [None]
        self.sv_input_access = ['item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _epw_file):

        try:
            import ladybug.epw as epw
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            epw_data = epw.EPW(_epw_file)
            location = epw_data.location
            dry_bulb_temperature = epw_data.dry_bulb_temperature
            dew_point_temperature = epw_data.dew_point_temperature
            relative_humidity = epw_data.relative_humidity
            wind_speed = epw_data.wind_speed
            wind_direction = epw_data.wind_direction
            direct_normal_rad = epw_data.direct_normal_radiation
            diffuse_horizontal_rad = epw_data.diffuse_horizontal_radiation
            global_horizontal_rad = epw_data.global_horizontal_radiation
            horizontal_infrared_rad = epw_data.horizontal_infrared_radiation_intensity
            direct_normal_ill = epw_data.direct_normal_illuminance
            diffuse_horizontal_ill = epw_data.diffuse_horizontal_illuminance
            global_horizontal_ill = epw_data.global_horizontal_illuminance
            total_sky_cover = epw_data.total_sky_cover
            barometric_pressure = epw_data.atmospheric_station_pressure
            model_year = epw_data.years
            g_temp = epw_data.monthly_ground_temperature
            ground_temperature = [g_temp[key] for key in sorted(g_temp.keys())]
        

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvImportEPW)

def unregister():
    bpy.utils.unregister_class(SvImportEPW)
