import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvImportSTAT(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvImportSTAT'
    bl_label = 'LB Import STAT'
    sv__stat_file: StringProperty(name='_stat_file', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_stat_file').prop_name = 'sv__stat_file'
        self.sv_output_names = ['location', 'ashrae_zone', 'koppen_zone', 'clear_dir_norm_rad', 'clear_diff_horiz_rad', 'ann_heat_dday_996', 'ann_heat_dday_990', 'ann_cool_dday_004', 'ann_cool_dday_010', 'monthly_ddays_050', 'monthly_ddays_100', 'extreme_cold_week', 'extreme_hot_week', 'typical_weeks']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['location', 'ashrae_zone', 'koppen_zone', 'clear_dir_norm_rad', 'clear_diff_horiz_rad', 'ann_heat_dday_996', 'ann_heat_dday_990', 'ann_cool_dday_004', 'ann_cool_dday_010', 'monthly_ddays_050', 'monthly_ddays_100', 'extreme_cold_week', 'extreme_hot_week', 'typical_weeks']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_stat_file']
        self.sv_input_types = ['string']
        self.sv_input_defaults = [None]
        self.sv_input_access = ['item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _stat_file):

        
        try:
            from ladybug.stat import STAT
            from ladybug.wea import Wea
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            stat_obj = STAT(_stat_file)
        
            # output location and climate zone
            location = stat_obj.location
            ashrae_zone = stat_obj.ashrae_climate_zone
            koppen_zone = stat_obj.koppen_climate_zone
        
            # output clear sky radiation
            try:  # first see if we can get the values from monthly optical depths
                wea = Wea.from_stat_file(_stat_file)
            except:  # no optical data was found; use the original clear sky
                wea = Wea.from_ashrae_clear_sky(location)
            clear_dir_norm_rad = wea.direct_normal_irradiance
            clear_diff_horiz_rad = wea.diffuse_horizontal_irradiance
        
            # output design day objects
            ann_heat_dday_996 = stat_obj.annual_heating_design_day_996
            ann_heat_dday_990 = stat_obj.annual_heating_design_day_990
            ann_cool_dday_004 = stat_obj.annual_cooling_design_day_004
            ann_cool_dday_010 = stat_obj.annual_cooling_design_day_010
            monthly_ddays_050 = stat_obj.monthly_cooling_design_days_050
            monthly_ddays_100 = stat_obj.monthly_cooling_design_days_100
        
            # output extreme and typical weeks
            extreme_cold_week = stat_obj.extreme_cold_week
            extreme_hot_week = stat_obj.extreme_hot_week
            typical_weeks = []
            seasonal_wks = [stat_obj.typical_winter_week, stat_obj.typical_spring_week,
                           stat_obj.typical_summer_week, stat_obj.typical_autumn_week]
            for wk in seasonal_wks:
                if wk is not None:
                    typical_weeks.append(wk)
            typical_weeks.extend(stat_obj.other_typical_weeks)

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvImportSTAT)

def unregister():
    bpy.utils.unregister_class(SvImportSTAT)
