import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvIncidentRadiation(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvIncidentRadiation'
    bl_label = 'LB Incident Radiation'
    sv__sky_mtx: StringProperty(name='_sky_mtx', update=updateNode)
    sv__geometry: StringProperty(name='_geometry', update=updateNode)
    sv_context_: StringProperty(name='context_', update=updateNode)
    sv__grid_size: StringProperty(name='_grid_size', update=updateNode)
    sv__offset_dist_: StringProperty(name='_offset_dist_', update=updateNode)
    sv_legend_par_: StringProperty(name='legend_par_', update=updateNode)
    sv_parallel_: StringProperty(name='parallel_', update=updateNode)
    sv__run: StringProperty(name='_run', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_sky_mtx').prop_name = 'sv__sky_mtx'
        self.inputs.new('SvStringsSocket', '_geometry').prop_name = 'sv__geometry'
        self.inputs.new('SvStringsSocket', 'context_').prop_name = 'sv_context_'
        self.inputs.new('SvStringsSocket', '_grid_size').prop_name = 'sv__grid_size'
        self.inputs.new('SvStringsSocket', '_offset_dist_').prop_name = 'sv__offset_dist_'
        self.inputs.new('SvStringsSocket', 'legend_par_').prop_name = 'sv_legend_par_'
        self.inputs.new('SvStringsSocket', 'parallel_').prop_name = 'sv_parallel_'
        self.inputs.new('SvStringsSocket', '_run').prop_name = 'sv__run'
        self.sv_output_names = ['points', 'results', 'total', 'mesh', 'legend', 'title', 'int_mtx']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['points', 'results', 'total', 'mesh', 'legend', 'title', 'int_mtx']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_sky_mtx', '_geometry', 'context_', '_grid_size', '_offset_dist_', 'legend_par_', 'parallel_', '_run']
        self.sv_input_types = ['System.Object', 'GeometryBase', 'GeometryBase', 'double', 'double', 'System.Object', 'bool', 'bool']
        self.sv_input_defaults = [None, None, None, None, None, None, None, None]
        self.sv_input_access = ['item', 'list', 'list', 'item', 'item', 'item', 'item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _sky_mtx, _geometry, context_, _grid_size, _offset_dist_, legend_par_, parallel_, _run):

        import math
        try:  # python 2
            from itertools import izip as zip
        except ImportError:  # python 3
            pass
        
        try:
            from ladybug.viewsphere import view_sphere
            from ladybug.graphic import GraphicContainer
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.config import conversion_to_meters
            from ladybug_tools.togeometry import to_joined_gridded_mesh3d
            from ladybug_tools.fromgeometry import from_mesh3d, from_point3d, from_vector3d
            from ladybug_tools.fromobjects import legend_objects
            from ladybug_tools.text import text_objects
            from ladybug_tools.intersect import join_geometry_to_mesh, intersect_mesh_rays
            from ladybug_tools.sverchok import all_required_inputs, hide_output, \
                show_output, objectify_output, de_objectify_output
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            # set the default offset distance
            _offset_dist_ = _offset_dist_ if _offset_dist_ is not None \
                else 0.1 / conversion_to_meters()
        
            # create the gridded mesh from the geometry
            study_mesh = to_joined_gridded_mesh3d(_geometry, _grid_size, _offset_dist_)
            points = [from_point3d(pt) for pt in study_mesh.face_centroids]
        
            if _run:  # run the entire study
                # hide the study points and mesh the geometry and context
                hide_output(ghenv.Component, 1)
                shade_mesh = join_geometry_to_mesh(_geometry + context_)
        
                # deconstruct the matrix and get the sky dome vectors
                mtx = de_objectify_output(_sky_mtx)
                total_sky_rad = [dir_rad + dif_rad for dir_rad, dif_rad in zip(mtx[1], mtx[2])]
                lb_vecs = view_sphere.tregenza_dome_vectors if len(total_sky_rad) == 145 \
                    else view_sphere.reinhart_dome_vectors
                if mtx[0][0] != 0:  # there is a north input for sky; rotate vectors
                    north_angle = math.radians(mtx[0][0])
                    lb_vecs = [vec.rotate_xy(north_angle) for vec in lb_vecs]
                sky_vecs = [from_vector3d(vec) for vec in lb_vecs]
        
                # intersect the rays with the mesh
                normals = [from_vector3d(vec) for vec in study_mesh.face_normals]
                int_matrix_init, angles = intersect_mesh_rays(
                    shade_mesh, points, sky_vecs, normals, parallel=parallel_)
        
                # compute the results
                results = []
                int_matrix = []
                for int_vals, angles in zip(int_matrix_init, angles):
                    pt_rel = [ival * math.cos(ang) for ival, ang in zip(int_vals, angles)]
                    int_matrix.append(pt_rel)
                    rad_result = sum(r * w for r, w in zip(pt_rel, total_sky_rad))
                    results.append(rad_result)
        
                # output the intersection matrix and compute total radiation
                int_mtx = objectify_output('Geometry/Sky Intersection Matrix', int_matrix)
                unit_conv = conversion_to_meters() ** 2
                total = 0
                for rad, area in zip(results, study_mesh.face_areas):
                    total += rad * area * unit_conv
        
                # create the mesh and legend outputs
                graphic = GraphicContainer(results, study_mesh.min, study_mesh.max, legend_par_)
                graphic.legend_parameters.title = 'kWh/m2'
                title = text_objects(
                    'Incident Radiation', graphic.lower_title_location,
                    graphic.legend_parameters.text_height * 1.5,
                    graphic.legend_parameters.font)
        
                # create all of the visual outputs
                study_mesh.colors = graphic.value_colors
                mesh = from_mesh3d(study_mesh)
                legend = legend_objects(graphic.legend)
        
            else:  # only show the points
                show_output(ghenv.Component, 1)
        

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvIncidentRadiation)

def unregister():
    bpy.utils.unregister_class(SvIncidentRadiation)
