import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvMonthlyChart(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvMonthlyChart'
    bl_label = 'LB Monthly Chart'
    sv__data: StringProperty(name='_data', update=updateNode)
    sv__base_pt_: StringProperty(name='_base_pt_', update=updateNode)
    sv__x_dim_: StringProperty(name='_x_dim_', update=updateNode)
    sv__y_dim_: StringProperty(name='_y_dim_', update=updateNode)
    sv_stack_: StringProperty(name='stack_', update=updateNode)
    sv_percentile_: StringProperty(name='percentile_', update=updateNode)
    sv_global_title_: StringProperty(name='global_title_', update=updateNode)
    sv_y_axis_title_: StringProperty(name='y_axis_title_', update=updateNode)
    sv_legend_par_: StringProperty(name='legend_par_', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_data').prop_name = 'sv__data'
        self.inputs.new('SvStringsSocket', '_base_pt_').prop_name = 'sv__base_pt_'
        self.inputs.new('SvStringsSocket', '_x_dim_').prop_name = 'sv__x_dim_'
        self.inputs.new('SvStringsSocket', '_y_dim_').prop_name = 'sv__y_dim_'
        self.inputs.new('SvStringsSocket', 'stack_').prop_name = 'sv_stack_'
        self.inputs.new('SvStringsSocket', 'percentile_').prop_name = 'sv_percentile_'
        self.inputs.new('SvStringsSocket', 'global_title_').prop_name = 'sv_global_title_'
        self.inputs.new('SvStringsSocket', 'y_axis_title_').prop_name = 'sv_y_axis_title_'
        self.inputs.new('SvStringsSocket', 'legend_par_').prop_name = 'sv_legend_par_'
        self.sv_output_names = ['data_mesh', 'data_lines', 'legend', 'borders', 'labels', 'y_title', 'title']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['data_mesh', 'data_lines', 'legend', 'borders', 'labels', 'y_title', 'title']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_data', '_base_pt_', '_x_dim_', '_y_dim_', 'stack_', 'percentile_', 'global_title_', 'y_axis_title_', 'legend_par_']
        self.sv_input_types = ['System.Object', 'Point3d', 'double', 'double', 'bool', 'double', 'string', 'string', 'System.Object']
        self.sv_input_defaults = [None, None, None, None, None, None, None, None, None]
        self.sv_input_access = ['list', 'item', 'item', 'item', 'item', 'item', 'item', 'list', 'list']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _data, _base_pt_, _x_dim_, _y_dim_, stack_, percentile_, global_title_, y_axis_title_, legend_par_):

        try:
            from ladybug.monthlychart import MonthlyChart
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_geometry.geometry2d.pointvector import Point2D
            from ladybug_geometry.geometry3d.pointvector import Point3D
            from ladybug_geometry.geometry3d.plane import Plane
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.config import conversion_to_meters
            from ladybug_tools.config import tolerance
            from ladybug_tools.togeometry import to_point2d
            from ladybug_tools.fromgeometry import from_mesh3d, from_mesh2d, \
                from_polyline2d, from_linesegment2d
            from ladybug_tools.text import text_objects
            from ladybug_tools.fromobjects import legend_objects
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            # set default values for the chart dimensions
            z_val = _base_pt_.Z if _base_pt_ is not None else 0
            z_val_tol = z_val + tolerance
            _base_pt_ = to_point2d(_base_pt_) if _base_pt_ is not None else Point2D()
            _x_dim_ = _x_dim_ if _x_dim_ is not None else 10.0 / conversion_to_meters()
            _y_dim_ = _y_dim_ if _y_dim_ is not None else 40.0 / conversion_to_meters()
            stack_ = stack_ if stack_ is not None else False
            percentile_ = percentile_ if percentile_ is not None else 34.0
            lpar = legend_par_[0] if len(legend_par_) != 0 else None
        
            # create the monthly chart object and get the main pieces of geometry
            month_chart = MonthlyChart(_data, lpar, _base_pt_, _x_dim_, _y_dim_,
                                       stack_, percentile_)
            if len(legend_par_) > 1:
                if legend_par_[1].min is not None:
                    month_chart.set_minimum_by_index(legend_par_[1].min, 1)
                if legend_par_[1].max is not None:
                    month_chart.set_maximum_by_index(legend_par_[1].max, 1)
        
            #  get the main pieces of geometry
            d_meshes = month_chart.data_meshes
            if d_meshes is not None:
                data_mesh = [from_mesh2d(msh, z_val_tol) for msh in d_meshes]
            d_lines = month_chart.data_polylines
            if d_lines is not None:
                data_lines = [from_polyline2d(lin, z_val_tol) for lin in d_lines]
            borders = [from_polyline2d(month_chart.chart_border, z_val)] + \
                    [from_linesegment2d(line, z_val) for line in month_chart.y_axis_lines] + \
                    [from_linesegment2d(line, z_val_tol) for line in month_chart.month_lines]
            legend = legend_objects(month_chart.legend)
        
            # process all of the text-related outputs
            title_txt = month_chart.title_text if global_title_ is None else global_title_
            txt_hgt = month_chart.legend_parameters.text_height
            font = month_chart.legend_parameters.font
            title = text_objects(title_txt, month_chart.lower_title_location, txt_hgt, font)
        
            # process the first y axis
            y1_txt = month_chart.y_axis_title_text1 if len(y_axis_title_) == 0 else y_axis_title_[0]
            y_title = text_objects(y1_txt, month_chart.y_axis_title_location1, txt_hgt, font)
            label1 = [text_objects(txt, Plane(o=Point3D(pt.x, pt.y, z_val)), txt_hgt, font, 1, 0)
                      for txt, pt in zip(month_chart.month_labels, month_chart.month_label_points)]
            label2 = [text_objects(txt, Plane(o=Point3D(pt.x, pt.y, z_val)), txt_hgt, font, 2, 3)
                      for txt, pt in zip(month_chart.y_axis_labels1, month_chart.y_axis_label_points1)]
            labels = label1 + label2
        
            # process the second y axis if it exists
            if month_chart.y_axis_title_text2 is not None:
                y2_txt = month_chart.y_axis_title_text2 if len(y_axis_title_) <= 1 else y_axis_title_[1]
                y_title2 = text_objects(y2_txt, month_chart.y_axis_title_location2, txt_hgt, font)
                y_title = [y_title, y_title2]
                label3 = [text_objects(txt, Plane(o=Point3D(pt.x, pt.y, z_val)), txt_hgt, font, 0, 3)
                         for txt, pt in zip(month_chart.y_axis_labels2, month_chart.y_axis_label_points2)]
                labels = labels + label3
        

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvMonthlyChart)

def unregister():
    bpy.utils.unregister_class(SvMonthlyChart)
