import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvPMV(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvPMV'
    bl_label = 'LB PMV Comfort'
    sv__air_temp: StringProperty(name='_air_temp', update=updateNode)
    sv__mrt_: StringProperty(name='_mrt_', update=updateNode)
    sv__rel_humid: StringProperty(name='_rel_humid', update=updateNode)
    sv__air_speed_: StringProperty(name='_air_speed_', update=updateNode)
    sv__met_rate_: StringProperty(name='_met_rate_', update=updateNode)
    sv__clothing_: StringProperty(name='_clothing_', update=updateNode)
    sv_pmv_par_: StringProperty(name='pmv_par_', update=updateNode)
    sv__run: StringProperty(name='_run', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_air_temp').prop_name = 'sv__air_temp'
        self.inputs.new('SvStringsSocket', '_mrt_').prop_name = 'sv__mrt_'
        self.inputs.new('SvStringsSocket', '_rel_humid').prop_name = 'sv__rel_humid'
        self.inputs.new('SvStringsSocket', '_air_speed_').prop_name = 'sv__air_speed_'
        self.inputs.new('SvStringsSocket', '_met_rate_').prop_name = 'sv__met_rate_'
        self.inputs.new('SvStringsSocket', '_clothing_').prop_name = 'sv__clothing_'
        self.inputs.new('SvStringsSocket', 'pmv_par_').prop_name = 'sv_pmv_par_'
        self.inputs.new('SvStringsSocket', '_run').prop_name = 'sv__run'
        self.sv_output_names = ['pmv', 'ppd', 'set', 'comfort', 'condition', 'heat_loss', 'comf_obj']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['pmv', 'ppd', 'set', 'comfort', 'condition', 'heat_loss', 'comf_obj']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_air_temp', '_mrt_', '_rel_humid', '_air_speed_', '_met_rate_', '_clothing_', 'pmv_par_', '_run']
        self.sv_input_types = ['System.Object', 'System.Object', 'System.Object', 'System.Object', 'System.Object', 'System.Object', 'System.Object', 'bool']
        self.sv_input_defaults = [None, None, None, None, None, None, None, None]
        self.sv_input_access = ['item', 'item', 'item', 'item', 'item', 'item', 'item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _air_temp, _mrt_, _rel_humid, _air_speed_, _met_rate_, _clothing_, pmv_par_, _run):

        try:
            from ladybug.psychrometrics import humid_ratio_from_db_rh
            from ladybug.datatype.temperature import Temperature
            from ladybug.datacollection import BaseCollection
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_comfort.collection.pmv import PMV
            from ladybug_comfort.parameter.pmv import PMVParameter
            from ladybug_comfort.pmv import predicted_mean_vote
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_comfort:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        def extract_collections(input_list):
            """Process inputs into collections and floats."""
            defaults = [None, _air_temp, 0.1, None, 1.1, 0.7]
            data_colls = []
            for i, input in enumerate(input_list):
                if input is None:
                    input_list[i] = defaults[i]
                elif isinstance(input, BaseCollection):
                    data_colls.append(input)
                else:
                    try:
                        input_list[i] = float(input)
                    except ValueError as e:
                        raise TypeError('input {} is not valid. Expected float or '
                                        'DataCollection. Got {}'.format(input, type(input)))
            return input_list, data_colls
        
        if all_required_inputs(ghenv.Component) and _run is True:
            # Process inputs and assign defaults.
            input_list = [_air_temp, _mrt_, _air_speed_, _rel_humid, _met_rate_, _clothing_]
            input, data_colls = extract_collections(input_list)
            pmv_par = pmv_par_ or PMVParameter()
            
            if data_colls == []:
                # The inputs are all individual values.
                pmv_result = predicted_mean_vote(input[0], float(input[1]), input[2],
                                                 input[3], input[4], input[5], 0,
                                                 pmv_par.still_air_threshold)
                pmv = pmv_result['pmv']
                ppd = pmv_result['ppd']
                set = pmv_result['set']
                hr = humid_ratio_from_db_rh(input[0], input[3])
                comfort = pmv_par.is_comfortable(pmv_result['ppd'], hr)
                condition = pmv_par.discomfort_reason(pmv_result['pmv'], pmv_result['ppd'])
                heat_loss = [pmv_result['heat_loss']['cond'],
                            pmv_result['heat_loss']['sweat'],
                            pmv_result['heat_loss']['res_l'],
                            pmv_result['heat_loss']['res_s'],
                            pmv_result['heat_loss']['rad'],
                            pmv_result['heat_loss']['conv']]
            else:
                # The inputs include Data Collections.
                if not isinstance(_air_temp, BaseCollection):
                    _air_temp = data_colls[0].get_aligned_collection(
                        float(_air_temp), Temperature(), 'C')
                
                comf_obj = PMV(_air_temp, _rel_humid, _mrt_, _air_speed_, _met_rate_,
                    _clothing_, 0, pmv_par)
                pmv = comf_obj.predicted_mean_vote
                ppd = comf_obj.percentage_people_dissatisfied
                set = comf_obj.standard_effective_temperature
                comfort = comf_obj.is_comfortable
                condition = comf_obj.discomfort_reason
                heat_loss = [comf_obj.heat_loss_conduction,
                            comf_obj.heat_loss_sweating,
                            comf_obj.heat_loss_latent_respiration,
                            comf_obj.heat_loss_dry_respiration,
                            comf_obj.heat_loss_radiation,
                            comf_obj.heat_loss_convection]

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvPMV)

def unregister():
    bpy.utils.unregister_class(SvPMV)
