import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvRTrad(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvRTrad'
    bl_label = 'LB Real Time Incident Radiation'
    sv__int_mtx: StringProperty(name='_int_mtx', update=updateNode)
    sv__sky_mtx: StringProperty(name='_sky_mtx', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_int_mtx').prop_name = 'sv__int_mtx'
        self.inputs.new('SvStringsSocket', '_sky_mtx').prop_name = 'sv__sky_mtx'
        self.sv_output_names = ['results']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['results']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_int_mtx', '_sky_mtx']
        self.sv_input_types = ['System.Object', 'System.Object']
        self.sv_input_defaults = [None, None]
        self.sv_input_access = ['item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _int_mtx, _sky_mtx):

        
        try:  # python 2
            from itertools import izip as zip
        except ImportError:  # python 3
            pass
        
        try:
            from ladybug_tools.sverchok import all_required_inputs, de_objectify_output
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            # deconstruct the matrices and get the total radiation from each patch
            int_mtx = de_objectify_output(_int_mtx)
            sky_mtx = de_objectify_output(_sky_mtx)
            total_sky_rad = [dirr + difr for dirr, difr in zip(sky_mtx[1], sky_mtx[2])]
        
            # compute the results
            results = []
            for pt_rel in int_mtx:
                results.append(sum(r * w for r, w in zip(pt_rel, total_sky_rad)))
        

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvRTrad)

def unregister():
    bpy.utils.unregister_class(SvRTrad)
