import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvComponentSolarMRT(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvComponentSolarMRT'
    bl_label = 'LB Solar MRT from Solar Components'
    sv__location: StringProperty(name='_location', update=updateNode)
    sv__longwave_mrt: StringProperty(name='_longwave_mrt', update=updateNode)
    sv__dir_horiz_rad: StringProperty(name='_dir_horiz_rad', update=updateNode)
    sv__diff_horiz_rad: StringProperty(name='_diff_horiz_rad', update=updateNode)
    sv_fract_body_exp_: StringProperty(name='fract_body_exp_', update=updateNode)
    sv__ground_ref_: StringProperty(name='_ground_ref_', update=updateNode)
    sv__solar_body_par_: StringProperty(name='_solar_body_par_', update=updateNode)
    sv__run: StringProperty(name='_run', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_location').prop_name = 'sv__location'
        self.inputs.new('SvStringsSocket', '_longwave_mrt').prop_name = 'sv__longwave_mrt'
        self.inputs.new('SvStringsSocket', '_dir_horiz_rad').prop_name = 'sv__dir_horiz_rad'
        self.inputs.new('SvStringsSocket', '_diff_horiz_rad').prop_name = 'sv__diff_horiz_rad'
        self.inputs.new('SvStringsSocket', 'fract_body_exp_').prop_name = 'sv_fract_body_exp_'
        self.inputs.new('SvStringsSocket', '_ground_ref_').prop_name = 'sv__ground_ref_'
        self.inputs.new('SvStringsSocket', '_solar_body_par_').prop_name = 'sv__solar_body_par_'
        self.inputs.new('SvStringsSocket', '_run').prop_name = 'sv__run'
        self.sv_output_names = ['erf', 'dmrt', 'mrt']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['erf', 'dmrt', 'mrt']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_location', '_longwave_mrt', '_dir_horiz_rad', '_diff_horiz_rad', 'fract_body_exp_', '_ground_ref_', '_solar_body_par_', '_run']
        self.sv_input_types = ['System.Object', 'System.Object', 'System.Object', 'System.Object', 'System.Object', 'System.Object', 'System.Object', 'bool']
        self.sv_input_defaults = [None, None, None, None, None, None, None, None]
        self.sv_input_access = ['item', 'item', 'item', 'item', 'item', 'item', 'item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _location, _longwave_mrt, _dir_horiz_rad, _diff_horiz_rad, fract_body_exp_, _ground_ref_, _solar_body_par_, _run):

        try:
            from ladybug_comfort.collection.solarcal import HorizontalSolarCal
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component) and _run is True:
            solar_mrt_obj = HorizontalSolarCal(_location, _dir_horiz_rad,
                                               _diff_horiz_rad, _longwave_mrt,
                                               fract_body_exp_, _ground_ref_, _solar_body_par_)
        
            erf = solar_mrt_obj.effective_radiant_field
            dmrt = solar_mrt_obj.mrt_delta
            mrt = solar_mrt_obj.mean_radiant_temperature

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvComponentSolarMRT)

def unregister():
    bpy.utils.unregister_class(SvComponentSolarMRT)
