import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvHeatmap(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvHeatmap'
    bl_label = 'LB Spatial Heatmap'
    sv__values: StringProperty(name='_values', update=updateNode)
    sv__mesh: StringProperty(name='_mesh', update=updateNode)
    sv_offset_dom_: StringProperty(name='offset_dom_', update=updateNode)
    sv_legend_par_: StringProperty(name='legend_par_', update=updateNode)
    sv_legend_title_: StringProperty(name='legend_title_', update=updateNode)
    sv_global_title_: StringProperty(name='global_title_', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_values').prop_name = 'sv__values'
        self.inputs.new('SvStringsSocket', '_mesh').prop_name = 'sv__mesh'
        self.inputs.new('SvStringsSocket', 'offset_dom_').prop_name = 'sv_offset_dom_'
        self.inputs.new('SvStringsSocket', 'legend_par_').prop_name = 'sv_legend_par_'
        self.inputs.new('SvStringsSocket', 'legend_title_').prop_name = 'sv_legend_title_'
        self.inputs.new('SvStringsSocket', 'global_title_').prop_name = 'sv_global_title_'
        self.sv_output_names = ['mesh', 'legend', 'title', 'colors', 'legend_par']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['mesh', 'legend', 'title', 'colors', 'legend_par']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_values', '_mesh', 'offset_dom_', 'legend_par_', 'legend_title_', 'global_title_']
        self.sv_input_types = ['double', 'Mesh', 'Interval', 'System.Object', 'string', 'string']
        self.sv_input_defaults = [None, None, None, None, None, None]
        self.sv_input_access = ['list', 'item', 'item', 'item', 'item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _values, _mesh, offset_dom_, legend_par_, legend_title_, global_title_):

        try:
            from ladybug.graphic import GraphicContainer
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.togeometry import to_mesh3d
            from ladybug_tools.fromgeometry import from_mesh3d
            from ladybug_tools.fromobjects import legend_objects
            from ladybug_tools.text import text_objects
            from ladybug_tools.color import color_to_color
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            # generate Ladybug objects
            lb_mesh = to_mesh3d(_mesh)
            if offset_dom_:
                lb_mesh = lb_mesh.height_field_mesh(
                    _values, (offset_dom_.T0, offset_dom_.T1))
            graphic = GraphicContainer(_values, lb_mesh.min, lb_mesh.max, legend_par_)
        
            # generate titles
            if legend_title_ is not None:
                graphic.legend_parameters.title = legend_title_
            if global_title_ is not None:
                title = text_objects(global_title_, graphic.lower_title_location,
                                     graphic.legend_parameters.text_height * 1.5,
                                     graphic.legend_parameters.font)
        
            # draw tools objects
            lb_mesh.colors = graphic.value_colors
            mesh = from_mesh3d(lb_mesh)
            legend = legend_objects(graphic.legend)
            colors = [color_to_color(col) for col in lb_mesh.colors]
            legend_par = graphic.legend_parameters
        

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvHeatmap)

def unregister():
    bpy.utils.unregister_class(SvHeatmap)
