import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvTimeOp(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvTimeOp'
    bl_label = 'LB Time Interval Operation'
    sv__data: StringProperty(name='_data', update=updateNode)
    sv__operation_: StringProperty(name='_operation_', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_data').prop_name = 'sv__data'
        self.inputs.new('SvStringsSocket', '_operation_').prop_name = 'sv__operation_'
        self.sv_output_names = ['daily', 'monthly', 'mon_per_hr']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['daily', 'monthly', 'mon_per_hr']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_data', '_operation_']
        self.sv_input_types = ['System.Object', 'string']
        self.sv_input_defaults = [None, None]
        self.sv_input_access = ['item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _data, _operation_):

        try:
            from ladybug.datacollection import HourlyDiscontinuousCollection
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.sverchok import all_required_inputs
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            assert isinstance(_data, HourlyDiscontinuousCollection), \
                '_data must be an Hourly Data Collection.' \
                ' Got {}.'.format(type(_data))
            if _operation_ is None:
                _operation_ = 'total' if _data.header.data_type.cumulative else 'average'
        
            if _operation_.lower() == 'average':
                daily = _data.average_daily()
                monthly = _data.average_monthly()
                mon_per_hr = _data.average_monthly_per_hour()
            elif _operation_.lower() == 'total':
                daily = _data.total_daily()
                monthly = _data.total_monthly()
                mon_per_hr = _data.total_monthly_per_hour()
            else:
                try:
                    percentile = float(_operation_)
                except ValueError as e:
                    raise TypeError(" Input '{}' for _operation_ is not valid. \n"
                                    "operation_ must be one of the following:\n"
                                    " average\n total\n [a number between 0 and "
                                    "100]".format(_operation_))
                else:
                    daily = _data.percentile_daily(percentile)
                    monthly = _data.percentile_monthly(percentile)
                    mon_per_hr = _data.percentile_monthly_per_hour(percentile)

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvTimeOp)

def unregister():
    bpy.utils.unregister_class(SvTimeOp)
