import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvViewPercent(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvViewPercent'
    bl_label = 'LB View Percent'
    sv__view_type: StringProperty(name='_view_type', update=updateNode)
    sv__resolution_: StringProperty(name='_resolution_', update=updateNode)
    sv__geometry: StringProperty(name='_geometry', update=updateNode)
    sv_context_: StringProperty(name='context_', update=updateNode)
    sv__grid_size: StringProperty(name='_grid_size', update=updateNode)
    sv__offset_dist_: StringProperty(name='_offset_dist_', update=updateNode)
    sv__geo_block_: StringProperty(name='_geo_block_', update=updateNode)
    sv_legend_par_: StringProperty(name='legend_par_', update=updateNode)
    sv_parallel_: StringProperty(name='parallel_', update=updateNode)
    sv__run: StringProperty(name='_run', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_view_type').prop_name = 'sv__view_type'
        self.inputs.new('SvStringsSocket', '_resolution_').prop_name = 'sv__resolution_'
        self.inputs.new('SvStringsSocket', '_geometry').prop_name = 'sv__geometry'
        self.inputs.new('SvStringsSocket', 'context_').prop_name = 'sv_context_'
        self.inputs.new('SvStringsSocket', '_grid_size').prop_name = 'sv__grid_size'
        self.inputs.new('SvStringsSocket', '_offset_dist_').prop_name = 'sv__offset_dist_'
        self.inputs.new('SvStringsSocket', '_geo_block_').prop_name = 'sv__geo_block_'
        self.inputs.new('SvStringsSocket', 'legend_par_').prop_name = 'sv_legend_par_'
        self.inputs.new('SvStringsSocket', 'parallel_').prop_name = 'sv_parallel_'
        self.inputs.new('SvStringsSocket', '_run').prop_name = 'sv__run'
        self.sv_output_names = ['points', 'view_vecs', 'results', 'mesh', 'legend', 'title', 'int_mtx']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['points', 'view_vecs', 'results', 'mesh', 'legend', 'title', 'int_mtx']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_view_type', '_resolution_', '_geometry', 'context_', '_grid_size', '_offset_dist_', '_geo_block_', 'legend_par_', 'parallel_', '_run']
        self.sv_input_types = ['string', 'int', 'GeometryBase', 'GeometryBase', 'double', 'double', 'bool', 'System.Object', 'bool', 'bool']
        self.sv_input_defaults = [None, None, None, None, None, None, None, None, None, None]
        self.sv_input_access = ['item', 'item', 'list', 'list', 'item', 'item', 'item', 'item', 'item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _view_type, _resolution_, _geometry, context_, _grid_size, _offset_dist_, _geo_block_, legend_par_, parallel_, _run):

        import math
        try:  # python 2
            from itertools import izip as zip
        except ImportError:  # python 3
            pass
        
        try:
            from ladybug.viewsphere import view_sphere
            from ladybug.color import Colorset
            from ladybug.graphic import GraphicContainer
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.config import conversion_to_meters
            from ladybug_tools.togeometry import to_joined_gridded_mesh3d, to_vector3d
            from ladybug_tools.fromgeometry import from_mesh3d, from_point3d, from_vector3d
            from ladybug_tools.fromobjects import legend_objects
            from ladybug_tools.text import text_objects
            from ladybug_tools.intersect import join_geometry_to_mesh, intersect_mesh_rays
            from ladybug_tools.sverchok import all_required_inputs, hide_output, \
                show_output, objectify_output
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        # dictionary to record all available view types
        VIEW_TYPES = {
            'HorizontalRadial': 'Horizontal Radial',
            'Horizonta30DegreeOffset': 'Horizontal 30-Degree Offset',
            'Spherical': 'Spherical',
            'SkyExposure': 'Sky Exposure',
            'SkyView': 'Sky View',
            '0': 'Horizontal Radial',
            '1': 'Horizontal 30-Degree Offset',
            '2': 'Spherical',
            '3': 'Sky Exposure',
            '4': 'Sky View'
        }
        
        
        if all_required_inputs(ghenv.Component):
            # process the view_type_ and set the default values
            vt_str = VIEW_TYPES[_view_type]
            _resolution_ = _resolution_ if _resolution_ is not None else 1
            _offset_dist_ = _offset_dist_ if _offset_dist_ is not None \
                else 0.1 / conversion_to_meters()
            if _geo_block_ is None:
                _geo_block_ = True if vt_str in ('Sky Exposure', 'Sky View') else False
        
            # create the gridded mesh from the geometry
            study_mesh = to_joined_gridded_mesh3d(_geometry, _grid_size, _offset_dist_)
            points = [from_point3d(pt) for pt in study_mesh.face_centroids]
        
            # get the view vectors based on the view type
            patch_wghts = None
            if vt_str == 'Horizontal Radial':
                lb_vecs = view_sphere.horizontal_radial_vectors(30 * _resolution_)
            elif vt_str == 'Horizontal 30-Degree Offset':
                patch_mesh, lb_vecs = view_sphere.horizontal_radial_patches(30, _resolution_)
                patch_wghts = view_sphere.horizontal_radial_patch_weights(30, _resolution_)
            elif vt_str == 'Spherical':
                patch_mesh, lb_vecs = view_sphere.sphere_patches(_resolution_)
                patch_wghts = view_sphere.sphere_patch_weights(_resolution_)
            else:
                patch_mesh, lb_vecs = view_sphere.dome_patches(_resolution_)
                patch_wghts = view_sphere.dome_patch_weights(_resolution_)
            view_vecs = [from_vector3d(pt) for pt in lb_vecs]
        
            if _run:  # run the entire study
                # hide the study points and mesh the geometry and context
                hide_output(ghenv.Component, 1)
                shade_mesh = join_geometry_to_mesh(_geometry + context_) if _geo_block_ \
                    else join_geometry_to_mesh(context_)
        
                # intersect the rays with the mesh
                if vt_str == 'Sky View':  # account for the normals of the surface
                    normals = [from_vector3d(vec) for vec in study_mesh.face_normals]
                    int_matrix, angles = intersect_mesh_rays(
                        shade_mesh, points, view_vecs, normals, parallel=parallel_)
                else:
                    int_matrix, angles = intersect_mesh_rays(
                        shade_mesh, points, view_vecs, parallel=parallel_)
        
                # compute the results
                int_mtx = objectify_output('View Intersection Matrix', int_matrix)
                vec_count = len(view_vecs)
                results = []
                if vt_str == 'Sky View':  # weight intersections by angle before output
                    for int_vals, angles in zip(int_matrix, angles):
                        w_res = (ival * 2 * math.cos(ang) for ival, ang in zip(int_vals, angles))
                        weight_result = sum(r * w for r, w in zip(w_res, patch_wghts))
                        results.append(weight_result * 100 / vec_count)
                else:
                    if patch_wghts:
                        for int_list in int_matrix:
                            weight_result = sum(r * w for r, w in zip(int_list, patch_wghts))
                            results.append(weight_result * 100 / vec_count)
                    else:
                        results = [sum(int_list) * 100 / vec_count for int_list in int_matrix]
        
                # create the mesh and legend outputs
                graphic = GraphicContainer(results, study_mesh.min, study_mesh.max, legend_par_)
                graphic.legend_parameters.title = '%'
                if legend_par_ is None or legend_par_.are_colors_default:
                    graphic.legend_parameters.colors = Colorset.view_study()
                title_txt = vt_str if vt_str in ('Sky Exposure', 'Sky View') else \
                    '{} View'.format(vt_str)
                title = text_objects(
                    title_txt, graphic.lower_title_location,
                    graphic.legend_parameters.text_height * 1.5,
                    graphic.legend_parameters.font)
        
                # create all of the visual outputs
                study_mesh.colors = graphic.value_colors
                mesh = from_mesh3d(study_mesh)
                legend = legend_objects(graphic.legend)
        
            else:  # only show the points and the view vectors
                show_output(ghenv.Component, 1)
        

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvViewPercent)

def unregister():
    bpy.utils.unregister_class(SvViewPercent)
