import bpy
import ladybug_tools.helper
from bpy.props import StringProperty
from sverchok.node_tree import SverchCustomTreeNode
from sverchok.data_structure import updateNode, zip_long_repeat

ghenv = ladybug_tools.helper.ghenv

class SvVisibilityPercent(bpy.types.Node, SverchCustomTreeNode):
    bl_idname = 'SvVisibilityPercent'
    bl_label = 'LB Visibility Percent'
    sv__view_points: StringProperty(name='_view_points', update=updateNode)
    sv_pt_weights_: StringProperty(name='pt_weights_', update=updateNode)
    sv__geometry: StringProperty(name='_geometry', update=updateNode)
    sv_context_: StringProperty(name='context_', update=updateNode)
    sv__grid_size: StringProperty(name='_grid_size', update=updateNode)
    sv__offset_dist_: StringProperty(name='_offset_dist_', update=updateNode)
    sv__geo_block_: StringProperty(name='_geo_block_', update=updateNode)
    sv_legend_par_: StringProperty(name='legend_par_', update=updateNode)
    sv_parallel_: StringProperty(name='parallel_', update=updateNode)
    sv__run: StringProperty(name='_run', update=updateNode)

    def sv_init(self, context):
        self.inputs.new('SvStringsSocket', '_view_points').prop_name = 'sv__view_points'
        self.inputs.new('SvStringsSocket', 'pt_weights_').prop_name = 'sv_pt_weights_'
        self.inputs.new('SvStringsSocket', '_geometry').prop_name = 'sv__geometry'
        self.inputs.new('SvStringsSocket', 'context_').prop_name = 'sv_context_'
        self.inputs.new('SvStringsSocket', '_grid_size').prop_name = 'sv__grid_size'
        self.inputs.new('SvStringsSocket', '_offset_dist_').prop_name = 'sv__offset_dist_'
        self.inputs.new('SvStringsSocket', '_geo_block_').prop_name = 'sv__geo_block_'
        self.inputs.new('SvStringsSocket', 'legend_par_').prop_name = 'sv_legend_par_'
        self.inputs.new('SvStringsSocket', 'parallel_').prop_name = 'sv_parallel_'
        self.inputs.new('SvStringsSocket', '_run').prop_name = 'sv__run'
        self.sv_output_names = ['points', 'results', 'mesh', 'legend', 'title', 'int_mtx']
        for name in self.sv_output_names:
            self.outputs.new('SvStringsSocket', name)

    def process(self):
        if not any(socket.is_linked for socket in self.outputs):
            return

        self.sv_output_names = ['points', 'results', 'mesh', 'legend', 'title', 'int_mtx']
        for name in self.sv_output_names:
            setattr(self, '{}_out'.format(name), [])
        self.sv_input_names = ['_view_points', 'pt_weights_', '_geometry', 'context_', '_grid_size', '_offset_dist_', '_geo_block_', 'legend_par_', 'parallel_', '_run']
        self.sv_input_types = ['Point3d', 'double', 'GeometryBase', 'GeometryBase', 'double', 'double', 'bool', 'System.Object', 'bool', 'bool']
        self.sv_input_defaults = [None, None, None, None, None, None, None, None, None, None]
        self.sv_input_access = ['list', 'list', 'list', 'list', 'item', 'item', 'item', 'item', 'item', 'item']
        sv_inputs_nested = []
        for name in self.sv_input_names:
            sv_inputs_nested.append(self.inputs[name].sv_get())
        for sv_input_nested in zip_long_repeat(*sv_inputs_nested):
            for sv_input in zip_long_repeat(*sv_input_nested):
                sv_input = list(sv_input)
                for i, value in enumerate(sv_input):
                    if self.sv_input_access[i] == 'list':
                        if isinstance(value, (list, tuple)):
                            values = value
                        else:
                            values = [value]
                        value = [self.sv_cast(v, self.sv_input_types[i], self.sv_input_defaults[i]) for v in values]
                        if value == [None]:
                            value = []
                        sv_input[i] = value
                    else:
                        sv_input[i] = self.sv_cast(value, self.sv_input_types[i], self.sv_input_defaults[i])
                self.process_ladybug(*sv_input)
        for name in self.sv_output_names:
            value = getattr(self, '{}_out'.format(name))
            # Not sure if this hack is correct, will find out when more nodes are generated
            #if len(value) == 0 or not isinstance(value[0], (list, tuple)):
            #    value = [value]
            self.outputs[name].sv_set(value)

    def sv_cast(self, value, data_type, default):
        result = default if isinstance(value, str) and value == '' else value
        if result is None and data_type == 'bool':
            return False
        return result

    def process_ladybug(self, _view_points, pt_weights_, _geometry, context_, _grid_size, _offset_dist_, _geo_block_, legend_par_, parallel_, _run):

        try:
            from ladybug.color import Colorset
            from ladybug.graphic import GraphicContainer
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_tools.config import conversion_to_meters
            from ladybug_tools.togeometry import to_joined_gridded_mesh3d, to_vector3d
            from ladybug_tools.fromgeometry import from_mesh3d, from_point3d, from_vector3d
            from ladybug_tools.fromobjects import legend_objects
            from ladybug_tools.text import text_objects
            from ladybug_tools.intersect import join_geometry_to_mesh, intersect_mesh_lines
            from ladybug_tools.sverchok import all_required_inputs, hide_output, \
                show_output, objectify_output
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_tools:\n\t{}'.format(e))
        
        
        if all_required_inputs(ghenv.Component):
            # set the default values
            _offset_dist_ = _offset_dist_ if _offset_dist_ is not None \
                else 0.1 / conversion_to_meters()
            if pt_weights_:
                assert len(pt_weights_) == len(_view_points), \
                    'The number of pt_weights_({}) must match the number of _view_points ' \
                    '({}).'.format(len(pt_weights_), len(_view_points))
        
            # create the gridded mesh from the geometry
            study_mesh = to_joined_gridded_mesh3d(_geometry, _grid_size, _offset_dist_)
            points = [from_point3d(pt) for pt in study_mesh.face_centroids]
        
            if _run:  # run the entire study
                # hide the study points and mesh the geometry and context
                hide_output(ghenv.Component, 1)
                shade_mesh = join_geometry_to_mesh(_geometry + context_) if _geo_block_ \
                    or _geo_block_ is None else join_geometry_to_mesh(context_)
        
                # intersect the lines with the mesh
                int_matrix = intersect_mesh_lines(
                    shade_mesh, points, _view_points, parallel=parallel_)
        
                # compute the results
                int_mtx = objectify_output('Visibility Intersection Matrix', int_matrix)
                vec_count = len(_view_points)
                if pt_weights_:  # weight intersections by the input point weights
                    tot_wght = sum(pt_weights_) / vec_count
                    adj_weights = [wght / tot_wght for wght in pt_weights_]
                    results = []
                    for int_vals in int_matrix:
                        w_res = [ival * wght for ival, wght in zip(int_vals, adj_weights)]
                        results.append(sum(w_res) * 100 / vec_count)
                else:  # no need to wieght results
                    results = [sum(int_list) * 100 / vec_count for int_list in int_matrix]
        
                # create the mesh and legend outputs
                graphic = GraphicContainer(results, study_mesh.min, study_mesh.max, legend_par_)
                graphic.legend_parameters.title = '%'
                if legend_par_ is None or legend_par_.are_colors_default:
                    graphic.legend_parameters.colors = Colorset.view_study()
                title = text_objects(
                    'Visibility Percent', graphic.lower_title_location,
                    graphic.legend_parameters.text_height * 1.5,
                    graphic.legend_parameters.font)
        
                # create all of the visual outputs
                study_mesh.colors = graphic.value_colors
                mesh = from_mesh3d(study_mesh)
                legend = legend_objects(graphic.legend)
        
            else:  # only show the points and the view vectors
                show_output(ghenv.Component, 1)
        

        for name in self.sv_output_names:
            if name in locals():
                getattr(self, '{}_out'.format(name)).append([locals()[name]])


def register():
    bpy.utils.register_class(SvVisibilityPercent)

def unregister():
    bpy.utils.unregister_class(SvVisibilityPercent)
